

      (* +------------------------------------------------------------+
         |  tude de quelques algorithmes pour le trac d'un segment  |
         +------------------------------------------------------------+ *)

(*
 * Auteur : Bruno Petazzoni (Bruno.Petazzoni@ac-idf.jussieu.fr).
 * Adapt  la bibliothque graphx par Michel Quercia (quercia@cal.enst.fr)
 * le 09/08/97.
 *)

#directory "..";; (* adapter en fonction des circonstances *)
open Graphx;;
#load "graphx.cmo";;

(*
 * Personalisation des fonctions graphiques : les fonctions suivantes
 * sont redfinies pour tracer des gros pixels carrs ou ronds.
 * L'implmentation est dfinie  la fin du fichier.
 *)

type dessinateur = {
  plot   : point -> unit;
  moveto : point -> unit;
  lineto : point -> unit;
  clear  : point -> unit;
  org    : point -> unit
};;

(* oprations sur les points *)

let ( ++ ) {x=x1; y=y1} {x=x2; y=y2} = {x=x1+x2; y=y1+y2}
and ( *. ) f {x=x; y=y} = {x=f*x; y=f*y}
;;

let nord       = {x=  0; y=  1}
and sud        = {x=  0; y= -1}
and est        = {x=  1; y=  0}
and ouest      = {x= -1; y=  0}
and nord_est   = {x=  1; y=  1}
and nord_ouest = {x= -1; y=  1}
and sud_est    = {x=  1; y= -1}
and sud_ouest  = {x= -1; y= -1}
;;

(* algorithme 1: on peint le pixel le plus haut sous le segment *)

let rec trace_segment_infrieur dess d p q =
 dess.plot p;
 if p.x=q.x then ()
 else begin
  let d = d+q.y in if d < q.x
    then trace_segment_infrieur dess d       (p ++ est)      q
    else trace_segment_infrieur dess (d-q.x) (p ++ nord_est) q
 end;;

(* algorithme 2: on peint le pixel le plus bas au-dessus du segment *)

let rec trace_segment_suprieur dess d p q =
 dess.plot p;
 if p.x = q.x then ()
 else begin
  let d = d-q.y in if d >= 0
    then trace_segment_suprieur dess d       (p ++ est)      q
    else trace_segment_suprieur dess (d+q.x) (p ++ nord_est) q
 end;;

(* algorithme 3: on peint le pixel le plus proche du segment *)

let rec trace_segment_mdian dess d p q =
 dess.plot p;
 if p.x = q.x then ()
 else begin
  let d = d+2*q.y in if d < 2*q.x
    then trace_segment_mdian dess d         (p ++ est)      q
    else trace_segment_mdian dess (d-2*q.x) (p ++ nord_est) q
 end;;

(* algorithme 4: on peint en faisant la moyenne des pixels voisins:
 les quatre coins pour Liss 1
 les milieux des cts en plus si Liss 2 *)

let test_pixel a c q eps =
 let b = q.y*eps.y-q.x*eps.x in
 if (b+a)*b <= c then 1 else 0;;

let make_palette l = let rec pal_of = function
 | [] -> []
 | t::q -> (0x101010*t)::(pal_of q)
 in Array.of_list (pal_of l);;

let lisse_pixel dess t p q =
 let d = 2*(q.y*p.x-q.x*p.y) in
 let a = 2*d and c = q.x*q.x+q.y*q.y-d*d in
 let f = test_pixel a c q in
 let points = match t with
  | 1 -> [sud_ouest; nord_ouest; sud_est; nord_est]
  | 2 -> [sud_ouest; nord_ouest; sud_est; nord_est; nord; sud; est; ouest]
  | _ -> failwith "taux de lissage inconnu"
 and palette = match t with
  | 1 -> make_palette [12;8;4;0]
  | 2 -> make_palette [14;12;10;8;6;4;2;0]
  | _ -> failwith "taux de lissage inconnu"
 in
 let lum = List.fold_left ( + ) 0 (List.map f points) in
 if lum>0 then
 begin
  set_color (palette.(lum-1));
  dess.plot p
 end;;

let rec trace_segment_liss dess t d p q =
 lisse_pixel dess t (p ++ 2*.sud)  q;
 lisse_pixel dess t (p ++ sud)     q;
 lisse_pixel dess t p              q;
 lisse_pixel dess t (p ++ nord)    q;
 lisse_pixel dess t (p ++ 2*.nord) q;
 if p.x = q.x then () else
 begin
  let d = d+2*q.y in if d < 2*q.x
    then trace_segment_liss dess t d         (p ++ est)      q
    else trace_segment_liss dess t (d-2*q.x) (p ++ nord_est) q
 end;;


(* affichage des exemples *)

type mode = Infrieur | Mdian | Suprieur | Liss of int;;

let do_exemple dess p mode =
 dess.clear p;
 begin
  match mode with
   | Infrieur -> trace_segment_infrieur dess 0   {x=0; y=0} p
   | Suprieur -> trace_segment_suprieur dess 0   {x=0; y=0} p
   | Mdian    -> trace_segment_mdian    dess p.x {x=0; y=0} p
   | Liss t   -> trace_segment_liss     dess t 0 {x=0; y=0} p
 end;
 set_color blue; dess.moveto {x=0; y=0}; dess.lineto p;
 dess.moveto {x=0; y=p.y+1};
 set_color black;
 let nom = match mode with
           | Infrieur -> "inf"
           | Suprieur -> "sup"
           | Mdian    -> "med"
           | Liss 1   -> "liss1"
           | Liss _   -> "liss2"
 in
 draw_string ("mode=" ^ nom)
;;

let exemple dess d =
  dess.org {x=0;        y=0}; 	  do_exemple dess d Infrieur;  
  dess.org {x=d.x+5;    y=0}; 	  do_exemple dess d Suprieur;  
  dess.org {x=2*d.x+10; y=0};     do_exemple dess d Mdian;  
  dess.org {x=0;        y=d.y+5}; do_exemple dess d (Liss 1);
  dess.org {x=d.x+5;    y=d.y+5}; do_exemple dess d (Liss 2)
;;

(* Implmentation des dessinateurs *)

type forme = Carr | Cercle;;

(*
 * forme  = forme des pixels
 * taille = taille des pixels
 * grille = affichage d'une grille (boolen)
 *)
let dessinateur forme taille grille =

  let orig = ref {x=0; y=0} in
  let scale(p) = taille*.(!orig ++ p ++ nord_est) in

  let mon_moveto p = moveto(scale(p) ++ (taille/2)*.nord_est)
  and mon_lineto p = lineto(scale(p) ++ (taille/2)*.nord_est)
  in

  let mon_plot p =
    let d = if grille then 1 else 0
    and f = taille/2  in
    match forme with
    | Carr  -> fill_rect   (scale(p) ++ d*.nord_est) (taille-d) (taille-d)
    | Cercle -> fill_circle (scale(p) ++ f*.nord_est) (f-d)

  and mon_clear p =
    if grille then begin
      set_color red;
      for i=0 to p.y do mon_moveto {x=0; y=i}; mon_lineto {x=p.x; y=i} done;
      for j=0 to p.x do mon_moveto {x=j; y=0}; mon_lineto {x=j; y=p.y} done
    end;
    set_color black

  and mon_org p = orig := p

  in {moveto = mon_moveto;
      lineto = mon_lineto;
      plot   = mon_plot;
      clear  = mon_clear;
      org    = mon_org}

;;


(* dmo *)
begin

  open_graph " 900x500";
  exemple (dessinateur Carr 13 true) {x=17; y=10};
  write_postscript "petazzoni.ps";
  moveto {x=10; y=450};
  draw_string "appuyer sur une touche ...";
  wait_next_event [Key_pressed];
  close_graph();
  print_endline "Le dessin est enregistr dans le fichier petazzoni.ps."

end;;

