(*******************************************************************************
 *                                                                             *
 *   Graphx : conversion vimage -> fig 3.2                                     *
 *                                                                             *
 *******************************************************************************)

open Graph;;
open Printf;;
open Utils;;

(* sort le code fig de [img] sur le canal [chout] *)
(* [comm] indique s'il faut placer en commentaire les commandes originales *)

let output_fig comm chout img =

  (* pb : xfig oriente l'axe des y vers le bas *)
  (* pour compenser cel, on calcule l'ordonne maximale dans la figure *)
  (* et on ajuste les ordonnes des points en consquence               *)
  let ymax = match bounding_box img with
  | VIDE -> 0
  | BB{ymax=ymax} -> ymax
  in

  let emit format = fprintf chout format
  and emitp format p = fprintf chout format p.x (ymax - p.y) in
  let emitv format v =
    for i=0 to Array.length v - 1 do emitp format v.(i) done;
    emit "\n"
  and comment = (if comm then comment chout "#" else fun _ -> ())
  in

  (* numrotation des couleurs *)
  let color = Hashtbl.create 17 in
  Hashtbl.add color black   0;	
  Hashtbl.add color white   7;	
  Hashtbl.add color red     4;	
  Hashtbl.add color green   2;	
  Hashtbl.add color blue    1;	
  Hashtbl.add color yellow  6;	
  Hashtbl.add color cyan    3;	
  Hashtbl.add color magenta 5;
  let colnum = ref 31 in
  let coul(c) =
    try Hashtbl.find color c
    with Not_found -> incr colnum; Hashtbl.add color c !colnum; !colnum
  in                      

  (* paramtres de trac *)
  let lw = ref 1   (* linewidth *)
  and co = ref(-1) (* color     *)
  in

  (* polygone *)
  let poly(fill,v) =
    emit "2 1 -1 %d %d %d 0 0 %d 0 0 0 0 0 0 %d\n"
         (if fill then 0 else !lw) !co !co (if fill then 20 else -1) (Array.length v);
    emitv " %d %d" v
  in

  (* courbe de Bzier *)
  (* pb : les cubiques de Bzier ne sont pas reprsentables dans le format *)
  (* FIG 3.2. On augmente le nombre de points de contrle par dichotomie   *)
  (* de faon  limiter la dissemblance.                                   *)
  let bezier(fill,v) =
    let v = poly_of_bezier 2 v in
    emit "3 0 -1 %d %d %d 0 0 %d 0 0 0 0 %d\n"
         (if fill then 0 else !lw) !co !co (if fill then 20 else -1) (Array.length v);
    emitv " %d %d" v;
    for i=0 to Array.length v -1 do
      emit " %d" (if i mod 3 = 0 then 0 else 1)
    done;
    emit "\n"
  in

  (* ellipse complte *)
  let ellipse(fill,p,rx,ry) =
    emit "1 1 -1 %d %d %d 0 0 %d 0 1 0"
         (if fill then 0 else !lw) !co !co (if fill then 20 else -1);
    emitp " %d %d"   p;
    emit  " %d %d"   rx ry;
    emitp " %d %d"   {x=p.x-rx; y=p.y-ry};
    emitp " %d %d\n" {x=p.x+rx; y=p.y+ry}
  in

  (* chaine de caractres *)
  (* il faut chapper les backslash et les caractres de code >= 128 *)
  let string(p,s) =
    emit "4 0 %d 0 0 %d %d 0 4 %d %d" !co (-1) 20 20 (10 * String.length(s));
    emitp " %d %d " p;
    for i=0 to String.length(s)-1 do match s.[i] with
    | ('\\' | '\128'..'\255' as c) ->
      let c = Char.code c in
      emit "\\%d%d%d" (c/64) ((c/8) land 7) (c land 7)
    | c -> emit "%c" c
    done;
    emit "\\001\n"
  in

  (* dtecte une couleur nouvelle et met sa dfinition *)
  let emit_color = function
  | Set_color(c) ->
    let oldnum = !colnum in
    let co = coul(c)     in
    if co > oldnum then emit "0 %d #%06x\n" co c
  | _ -> ()
  in

  (* fonction gnrale de conversion *)
  let fig(cmd) =
    comment cmd;
    match cmd with

    | Set_color(c)      -> co := coul(c)
    | Set_line_width(w) -> lw := w
    | Set_font(f)       -> ()
    | Set_text_size(sz) -> ()

    | Poly(fill,v) -> poly(fill,v)
    | Bezier(fill,d,v) ->
       if d >= 4 then bezier(fill,v) else poly(fill,poly_of_bezier d v)
    | Arc(fill,p,rx,ry,a1,a2) -> bezier(fill,bezier_of_arc(fill,p,rx,ry,a1,a2))
    | Ellipse(fill,p,rx,ry) -> ellipse(fill,p,rx,ry)
    | String(p,s) -> string(p,s)
  in

  (* prologue *)
  emit "#FIG 3.2\nPortrait\nFlush left\nMetric\nA4\n100.00\nSingle\n0\n100 2\n";

  (* parcourt la liste  blanc pour reprer les couleurs non prdfinies *)
  List.iter emit_color img;

  (* mmet toutes les commandes *)
  List.iter fig img;

  (* pilogue *)
  ()
;;
