(*****************************************************************************
 *                                                                           *
 *      Interface graphique : adaptation de la bibliothque libgraph         *
 *                                                                           *
 *****************************************************************************)

type point = {x:int; y:int};;
type color = int;;
type image = Graphics.image;;
type graphic_cmd =
| Set_color    	 of color
| Set_line_width of int
| Set_font       of string
| Set_text_size  of int
| Poly    	 of bool * (point array)
| Bezier 	 of bool * int * (point array)
| Arc     	 of bool * point * int * int * int * int
| Ellipse 	 of bool * point * int * int
| String  	 of point * string
;;
type vimage = graphic_cmd list;;
type status = {
  mouse      : point;         (* position de la souris                      *)
  button     : bool;          (* true si un bouton est enfonc              *)
  keypressed : bool;          (* true si une nouvelle touche est disponible *)
  key        : char }         (* le cractrre frapp                        *)

  and event =
  | Button_down               (* Appui sur un bouton                        *)
  | Button_up                 (* Relchement d'un bouton                    *)
  | Key_pressed               (* Appui sur une touche                       *)
  | Mouse_motion              (* Mouvement de la souris                     *)
  | Poll                      (* Ne pas attendre, retourner immdiatement   *)
;;
type bounding_box = VIDE | BB of bb
and bb = {xmin:int; xmax:int; ymin:int; ymax:int};;

                          (* +---------------------+
                             |  Courbes de Bzier  |
                             +---------------------+ *)


let milieu {x=a;y=b} {x=c;y=d} = {x=(a+c+1) asr 1; y=(b+d+1) asr 1};;
let poly_of_bezier d v =

  (* contrle le nombre de points fournis et alloue le tableau rsultat *)
  let lv = Array.length v in
  if (lv mod 3) <> 1 then raise(Invalid_argument "poly_of_bezier");
  let lr = (lv - 1) lsl d in
  let r = Array.create (lr+1) {x=0; y=0} in

  (* distribue les points initiaux : les coordonnes sont multiplies par *)
  (* 2^16 pour amliorer la prcision dans le calcul des milieux.         *)
  for i=0 to lv-1 do
    let {x=x;y=y} = v.(i) in r.(i) <- {x=x lsl 16; y=y lsl 16}
  done;

  (* divise les arcs [d] fois de suite *)
  let l = ref(lv-1) in
  while !l < lr do
    let i = ref !l and j = ref(!l lsl 1) in
    l := !j;
    while !i > 0 do
      let a = r.(!i) and b = r.(!i-1) and c = r.(!i-2) and d = r.(!i-3)  in
      let ab   = milieu a b   and bc  = milieu b c   and cd = milieu c d in
      let abc  = milieu ab bc and bcd = milieu bc cd in
      let abcd = milieu abc bcd in
      r.(!j) <- a;      r.(!j-1) <- ab;  r.(!j-2) <- abc;
      r.(!j-3) <- abcd; r.(!j-4) <- bcd; r.(!j-5) <- cd;
      i := !i-3; j := !j-6
      (* rmq : inutile de recopier le 1er point, [a], il sera mis en place   *)
      (* lors de l'itration suivante ou il est dj en place si c'est r.(0) *)
    done
  done;

  (* divise les points rsultant par 2^16 pour revenir  l'chelle initiale *)
  let c = 1 lsl 15 in
  for i=0 to lr do
    let {x=x;y=y} = r.(i) in r.(i) <- {x = (x+c) asr 16; y = (y+c) asr 16}
  done;
  r
;;



               (* +------------------------------------------+
                  |  Adaptation des fonctions de [libgraph]  |
                  +------------------------------------------+ *)

(*
 * Liste des commandes graphiques enregistres
 * L'enregistrement dans cette liste est ralis par [push] et control par
 * [vrecord].
 *)

let vimage  = ref []
and vrecord = ref true;;
let push cmd = if !vrecord then vimage := cmd :: !vimage;;


let open_graph arg   = if !vrecord then vimage := []; Graphics.open_graph arg
and close_graph      = Graphics.close_graph
and clear_graph ()   = if !vrecord then vimage := []; Graphics.clear_graph()
and size_x     	     = Graphics.size_x
and size_y     	     = Graphics.size_y
and rgb        	     = Graphics.rgb
and black      	     = Graphics.black  
and white      	     = Graphics.white  
and red        	     = Graphics.red    
and green      	     = Graphics.green  
and blue       	     = Graphics.blue   
and yellow     	     = Graphics.yellow 
and cyan       	     = Graphics.cyan   
and magenta    	     = Graphics.magenta
and background 	     = Graphics.background
and foreground 	     = Graphics.foreground
and point_color p    = Graphics.point_color p.x p.y
and moveto p         = Graphics.moveto p.x p.y
and current_point()  = let (x,y) = Graphics.current_point() in {x=x; y=y}
and set_color c      = push (Set_color c);      Graphics.set_color c
and set_line_width w = push (Set_line_width w); Graphics.set_line_width w
and set_font f       = push (Set_font (f^""));  Graphics.set_font f
and set_text_size sz = push (Set_text_size sz); Graphics.set_text_size sz
and text_size        = Graphics.text_size
;;

let arc fill p rx ry a1 a2 =
  push (Arc(fill,p,rx,ry,a1,a2));
  (if fill then Graphics.fill_arc else Graphics.draw_arc)
     p.x p.y rx ry a1 a2;;

let ellipse fill p rx ry =
  push (Ellipse(fill,p,rx,ry));
  (if fill then Graphics.fill_ellipse else Graphics.draw_ellipse)
    p.x p.y rx ry;;

let poly fill v =
  push (Poly(fill,Array.copy v));
  if fill then Graphics.fill_poly(Array.map (fun p -> (p.x,p.y)) v)
  else if Array.length v > 0 then begin
    let {x=x;y=y}=v.(0) in Graphics.moveto x y;
    for i=1 to (Array.length v - 1) do
      let {x=x;y=y}=v.(i) in Graphics.lineto x y
    done
  end;;

let bezier fill d v =
  push (Bezier(fill,d,Array.copy v));
  let w = poly_of_bezier d v in
  if fill then Graphics.fill_poly(Array.map (fun p -> (p.x,p.y)) w)
  else if Array.length w > 0 then begin
    let {x=x;y=y}=w.(0) in Graphics.moveto x y;
    for i=1 to (Array.length w - 1) do
      let {x=x;y=y}=w.(i) in Graphics.lineto x y
    done
  end;;

let draw_string s =
  push (String(current_point(),(s^"")));
  Graphics.draw_string s
;;

let draw_arc        = arc     false and fill_arc     = arc     true
and draw_ellipse    = ellipse false and fill_ellipse = ellipse true
and draw_poly       = poly    false and fill_poly    = poly    true
and draw_bezier     = bezier  false and fill_bezier  = bezier  true
;;

let plot p          = ellipse true p 1 1; moveto p
and lineto p        = poly false [|current_point(); p|]
and fill_rect p w h = poly true  [|p; {x=p.x+w;y=p.y}; {x=p.x+w;y=p.y+h}; {x=p.x;y=p.y+h}|]
and curveto b c d   = bezier false 4 [|current_point(); b; c; d|]
and draw_circle p r = ellipse false p r r
and fill_circle p r = ellipse true  p r r
and draw_char c     = draw_string (String.make 1 c)
;;

let convert_event = function
| Button_down  -> Graphics.Button_down
| Button_up    -> Graphics.Button_up
| Key_pressed  -> Graphics.Key_pressed
| Mouse_motion -> Graphics.Mouse_motion
| Poll         -> Graphics.Poll
and convert_status = function {
  Graphics.mouse_x    = x;
  Graphics.mouse_y    = y;
  Graphics.button     = b;
  Graphics.keypressed = p;
  Graphics.key        = k} -> {
  mouse      = {x=x; y=y};
  button     = b;
  keypressed = p;
  key        = k}
;;

let wait_next_event(list) =
  convert_status(Graphics.wait_next_event(List.map convert_event list))
;;

let mouse_pos()   = (wait_next_event[Poll]).mouse
and button_down() = (wait_next_event[Poll]).button
and read_key()    = (wait_next_event[Poll]).key
and key_pressed() = (wait_next_event[Poll]).keypressed
and sound         = Graphics.sound
;;

                       (* +-----------------+
                          |  Images bitmap  |
                          +-----------------+ *)

let transp	     = Graphics.transp
and make_image	     = Graphics.make_image
and dump_image	     = Graphics.dump_image
and draw_image	     = Graphics.draw_image
and get_image	     = Graphics.get_image
and create_image     = Graphics.create_image
and blit_image	     = Graphics.blit_image
;;

let current_image() = dump_image(get_image 0 0 (size_x()) (size_y()));;

                         (* +-----------------------+
                            |  Images vectorielles  |
                            +-----------------------+ *)

let current_vimage() = List.rev(!vimage)
and set_record(bool) = vrecord := bool
;;

let display = function
| Set_color(c)             -> set_color c           
| Set_line_width(w)	   -> set_line_width w
| Set_font(f)		   -> set_font f            
| Set_text_size(sz)	   -> set_text_size sz
| Poly(b,v)		   -> poly b v           
| Bezier(b,d,v)	           -> bezier b d v           
| Arc(b,p,rx,ry,a1,a2)     -> arc b p rx ry a1 a2
| Ellipse(b,p,rx,ry)	   -> ellipse b p rx ry  
| String(p,s)	           -> moveto p; draw_string s
;;

let display_vimage img = List.iter display img
;;

(* Calcul du rectangle d'encombrement *)

let bounding_box img =

  let pos = ref{x=0; y=0}  (* position courante    *)
  and lw  = ref 1          (* 1/2 largeur de trait *)
  and sz  = ref 20         (* text_size            *)
  in

  let rec encombrement = function

    Set_color(c) ->      VIDE
  | Set_line_width(w) -> lw := (w+1)/2; VIDE
  | Set_font(f) ->       VIDE
  | Set_text_size(s) ->  sz := s; VIDE

  | Poly(fill,v) -> 
      if (Array.length v) = 0 then VIDE
      else begin
        let xmin = ref v.(0).x and xmax = ref v.(0).x
        and ymin = ref v.(0).y and ymax = ref v.(0).y in
	for i=1 to (Array.length v) - 1 do
	  xmin := min !xmin v.(i).x; xmax := max !xmax v.(i).x;
	  ymin := min !ymin v.(i).y; ymax := max !ymax v.(i).y
	done;
        if fill then BB {xmin =	!xmin;       xmax = !xmax;
                         ymin =	!ymin;       ymax = !ymax}
                else BB {xmin =	!xmin - !lw; xmax = !xmax + !lw;
                         ymin =	!ymin - !lw; ymax = !ymax + !lw}
      end

  | Bezier(fill,d,v) -> encombrement(Poly(fill,v))

  | Arc(fill,p,a,b,u,v) ->
      let u' = (float u) *. 0.017453293
      and v' = (float v) *. 0.017453293
      and a' = float a
      and b' = float b in
      let x0 = truncate( a' *. cos(u'))
      and y0 = truncate( b' *. sin(u'))
      and x1 = truncate( a' *. cos(v'))
      and y1 = truncate( a' *. sin(v')) in

      let f(x) = if x >= 0 then x/360 else -(1-x/360) in
      
      let xmin = if f(u-180) = f(v-180) then min x0 x1 else (-a)
      and xmax = if f(u)     = f(v)     then max x0 x1 else a
      and ymin = if f(u+90)  = f(v+90)  then min y0 y1 else (-b)
      and ymax = if f(u-90)  = f(v-90)  then max y0 y1 else b
      in

      if fill then BB {xmin = p.x + xmin;       xmax = p.x + xmax;
                       ymin = p.y + ymin;       ymax = p.y + ymax}
              else BB {xmin = p.x + xmin - !lw; xmax = p.x + xmax + !lw;
                       ymin = p.y + ymin - !lw; ymax = p.y + ymax + !lw}

  | Ellipse(fill,p,a,b) ->
     if fill then BB {xmin = p.x - a;       xmax = p.x + a;
                      ymin = p.y - b;       ymax = p.y + b}
             else BB {xmin = p.x - a - !lw; xmax = p.x + a + !lw;
                      ymin = p.y - b - !lw; ymax = p.y + b + !lw}

  | String(p,s) ->
     if s = "" then VIDE
               else BB {xmin=p.x; xmax=p.x + !sz*(String.length s)/2;
                        ymin=p.y; ymax=p.y + !sz}

  in (* fin de la dfinition de [encombrement] *)

  let r = ref VIDE in
  let job(o) = match !r,encombrement(o) with
     VIDE,x -> r := x
  |  _,VIDE -> ()
  |  BB(a),BB(b) -> r := BB {xmin = min a.xmin b.xmin; xmax = max a.xmax b.xmax;
                             ymin = min a.ymin b.ymin; ymax = max a.ymax b.ymax}
  in
  List.iter job img;
  !r
;;
